// ------------------------------------------------------------------
// NetTacToe Full demonstration of CDXConnection
// ------------------------------------------------------------------
// Author : Jimb Esser (wasteland@writeme.com)
// Date : 1-28-1
// Version : 1.0

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#define CDXINCLUDEALL     // this define includes all headers, otherwise include one by one
#include <cdx.h>

#include "CDXConnection.h"
#include "addons/CDXVertTextScroller.h"
#include "TTTServer.h"
#include "TicTacToe.h"
#include "NetTacToe.h"

#include "resource.h"


// In Dialog.cpp
void GetValue(CDXScreen * Screen, HWND hWnd, HINSTANCE hInst,
			   char sName[], char sDefaultValue[],
			   char *sReturnVal); // Calls a dialog, and asks for sName
			   // sReturnVal should point to a char[256]




#define BOX_WIDTH 100
#define BOX_HEIGHT 100

#define STATE_SPLASH 0
#define STATE_MAIN 1
#define STATE_EXIT 2

int g_iGameState=STATE_SPLASH;
int g_iPlayerNum=-1;
CDXConnection g_con;
CTicTacToe g_ttt;
DWORD FgColor=0, BgColor=0;

bool bMouseClicked=false;
int MouseXPos;
int MouseYPos;

// ------------------------------------------------------------------
// Global Variables
// ------------------------------------------------------------------
char        szAppName[]     = "NetTacToe";
char        szClassName[]   = "NetTacToeWndClass";

HINSTANCE   g_hInst;			        // instance handle
HWND        g_hWnd;				        // window handle


// ------------------------------------------------------------------
// CDX Objects
// ------------------------------------------------------------------
CDXScreen   * Screen    = 0;	    // The screen object, every program must have one
CDXSurface  * Splash    = 0;        // splash screen surface
CDXInput    * Input     = 0;        // The input object
CDXVertTextScroller	* g_MessageBox = 0;


// ------------------------------------------------------------------
// cdx_Init - handles initialization of the CDX objects
// ------------------------------------------------------------------
BOOL cdx_Init()
{
	// Create the CDXSreen object
	Screen = new CDXScreen();
	if (Screen==NULL)
        CDXError( NULL , "Could not create CDXScreen object" );

	// start app in a window
	if (FAILED(Screen->CreateWindowed(g_hWnd, 640, 480)))
        CDXError( Screen , "Could not set windowed mode 640x480" );

	// load the splash screen
	Splash = new CDXSurface( );
    if(FAILED(Splash->Create( Screen , "SPLASH.BMP" )))
        CDXError( Screen , "Could not load file SPLASH into surface" );

	// create our input object
	Input = new CDXInput( );
	if( Input  ==  NULL ) 
        CDXError( Screen , "Could not create CDXInput object" );
	Input->SetActiveDevices(FALSE, TRUE, FALSE);

	// Create input devices
	if(FAILED(Input->Create( g_hInst , g_hWnd ))) 
        CDXError( Screen , "Could not create direct input object" );

	g_MessageBox = NULL;

	
	return TRUE;
}

void immediateRefresh() {
	g_MessageBox->DrawTrans(Screen->GetFront());
}

void addText(char *str, bool bImmediateRefresh) {
	int iIYinit=180;
	int iLen=strlen(str)+1;
	if (g_MessageBox!=NULL) {
		iIYinit = g_MessageBox->IY;
		iLen += strlen(g_MessageBox->Text)+1;
	}
	char *newtext = new char[iLen];
	if (g_MessageBox!=NULL) {
		strcpy(newtext, g_MessageBox->Text);
		delete g_MessageBox;
		g_MessageBox = NULL;
	} else {
		strcpy(newtext, "");
	}
	strcat(newtext, str);
	RECT ClipRectangle;
	ClipRectangle.left = 0;
	ClipRectangle.top = BOX_HEIGHT *3;
	ClipRectangle.right = 640;
	ClipRectangle.bottom = 480;
	g_MessageBox = new CDXVertTextScroller( Screen, &ClipRectangle, newtext);
	g_MessageBox->IY = iIYinit;
	if (str[strlen(str)-1]=='\n') {
		g_MessageBox->ScrollUp(16);
	}
	if (bImmediateRefresh) {
		immediateRefresh();
	}
}


// ------------------------------------------------------------------
// cdx_DeInit - handles cleanup of CDX objects
// ------------------------------------------------------------------
void cdx_DeInit( void )
{
	SAFEDELETE( g_MessageBox );
	SAFEDELETE( Splash );
	SAFEDELETE( Input );	
	SAFEDELETE( Screen );
}



void handleSplash() {
	// clear the current frame
	Screen->GetBack()->Fill(0);

	// handle keyboard input	

	// if Escape is pressed, close the application
	if (Input->GetKeyState(CDXKEY_ESCAPE) == CDXKEY_RELEASE)
	{
		SendMessage(g_hWnd, WM_CLOSE, 0, 0);
		// return out as the code below is invalid now
		g_iGameState = STATE_EXIT;
		return;
	}

	if (Input->GetKeyState(CDXKEY_S) == CDXKEY_RELEASE)
	{
		// Start server
		addText("Starting server.\n", true);
		startServer();
		// And start a client connecting to 127.0.0.1
		addText("Connecting to self...\n", true);
		if (FAILED(g_con.CreateClient())) {
			addText("Error creating client\n");
		} else if (FAILED(g_con.Connect("127.0.0.1", TTT_PORT))) {
			addText("Error conection to self as server\n");
		} else {
			addText("Connected.\n");
			g_iPlayerNum = g_con.getNodeNum();
			g_iGameState = STATE_MAIN;
		}
	}
	if (Input->GetKeyState(CDXKEY_C) == CDXKEY_RELEASE)
	{
		// Ask for IP and then Start client
		char destIP[256];
		GetValue(Screen, g_hWnd, g_hInst,
			   "Server IP:", "127.0.0.1",
			   destIP); // Calls a dialog, and asks for sName
			   // sReturnVal should point to a char[256]

		addText("Trying to connect...", true);
		if (FAILED(g_con.CreateClient())) {
			addText("Error creating client\n");
		} else if (FAILED(g_con.Connect(destIP, TTT_PORT))) {
			addText("Error conection to server at that IP\n");
		} else {
			addText("Connected.\n");
			g_iPlayerNum = g_con.getNodeNum();
			g_iGameState = STATE_MAIN;
		}
	}

	// draw the splash screen
	Splash->DrawBlk(Screen->GetBack(), 0, 0);
	if (FgColor==BgColor) {
		// Grab the colors  from the splash screen
		Splash->Lock();
		FgColor = Splash->GetPixel(1, 1);
		BgColor = Splash->GetPixel(2, 1);
		Splash->UnLock();
	}

	// Display messages
	if (g_MessageBox) {
		g_MessageBox->DrawTrans(Screen->GetBack());
	}

}

void drawBoard() {
	Screen->GetBack()->Line(BOX_WIDTH, 0, BOX_WIDTH, BOX_HEIGHT*3, FgColor);
	Screen->GetBack()->Line(BOX_WIDTH*2, 0, BOX_WIDTH*2, BOX_HEIGHT*3, FgColor);
	Screen->GetBack()->Line(0, BOX_HEIGHT, BOX_WIDTH*3, BOX_HEIGHT, FgColor);
	Screen->GetBack()->Line(0, BOX_HEIGHT*2, BOX_WIDTH*3, BOX_HEIGHT*2, FgColor);

	int xcenter, ycenter, rad, innerrad;
	for (int x=0; x<3; x++) {
		for (int y=0; y<3; y++) {
			switch (g_ttt.getMark(x, y)) {
			case 1:
				// X
				Screen->GetBack()->Line(x*BOX_WIDTH+10, y*BOX_HEIGHT+10,
					(x+1)*BOX_WIDTH-10, (y+1)*BOX_HEIGHT-10, FgColor);
				Screen->GetBack()->Line((x+1)*BOX_WIDTH-10, y*BOX_HEIGHT+10,
					x*BOX_WIDTH+10, (y+1)*BOX_HEIGHT-10, FgColor);
				break;
			case 2:
				// O
				xcenter = x*BOX_WIDTH + BOX_WIDTH / 2;
				ycenter = y*BOX_HEIGHT + BOX_HEIGHT / 2;
				rad = min(BOX_HEIGHT, BOX_WIDTH) / 2 - 10;
				innerrad = max(10, rad - 20);
				Screen->GetBack()->FillCircle(xcenter, ycenter, rad, FgColor);
				Screen->GetBack()->FillCircle(xcenter, ycenter, innerrad, BgColor);
				break;
			}
		}
	}

}

void handleMain() {
	// clear the current frame
	Screen->GetBack()->Fill(BgColor);

	// handle keyboard input	

	// if Escape is pressed, close the application
	if (Input->GetKeyState(CDXKEY_ESCAPE))
	{
		g_iGameState = STATE_EXIT;
		SendMessage(g_hWnd, WM_CLOSE, 0, 0);
		// return out as the code below is invalid now
		return;
	}

	// Check for click
	if (bMouseClicked) {
		bMouseClicked=false;
		int x = MouseXPos;
		int y = MouseYPos;
		char buf[255];
		sprintf(buf, "(%d, %d)", x, y);
		Screen->GetBack()->TextXY(10,10,RGB(0xff,0xff, 0xff), buf);
		x /= BOX_WIDTH;
		y /= BOX_HEIGHT;
		if ((x>=0) && (x<3) && (y>=0) && (y<3)) {
			if (g_ttt.isEmpty(x, y)) {
				char data[2];
				data[0] = (char)x;
				data[1] = (char)y;
				if (FAILED(g_con.sendData(NODE_SERVER, data, 2, TAG_MARK))) {
					addText("Error sending data to server.  Probably shut down.\n");
				}
				OutputDebugString(buf);
				// Send info to server
			}
		}
	}
	
	// if arrows are pressed scroll
	if (Input->GetKeyState(CDXKEY_UPARROW))
	{
		if (g_MessageBox) {
			g_MessageBox->ScrollUp(3);
		}
	}
	if (Input->GetKeyState(CDXKEY_DOWNARROW))
	{
		if (g_MessageBox) {
			g_MessageBox->ScrollDown(3);
		}
	}

	// Check for desire to chat
	if (Input->GetKeyState(CDXKEY_T)==CDXKEY_RELEASE)
	{
		char sMsg[256];
		GetValue(Screen, g_hWnd, g_hInst, "Message:", "", sMsg);
		char sMsg2[262];
		sprintf(sMsg2, "[%s]: %s\n", (g_iPlayerNum==0)?"X":"O", sMsg);
		if (FAILED(g_con.sendData(NODE_SERVER, sMsg2, strlen(sMsg2) + 1, TAG_CHAT))) {
			addText("Error sending data to server.  Probably shut down.\n");
		}	
	}

	// Check for incoming data and handle
	while (g_con.isDataReady()) {
		char data[1024];
		int iNodeNum, iNumRead;
		unsigned char iTag;
		g_con.getData(iNodeNum, data, 1024, iNumRead, &iTag);
		switch (iTag) {
		case TAG_CHAT:
			// Add to message box
			addText(data);
			break;
		case TAG_MARK:
			g_ttt.mark(data[0], data[1], data[2]);
			break;
		case TAG_ENDGAME:
			// Display win message
			if (data[0]==g_iPlayerNum+1) {
				addText("You won!\n");
			} else if (data[0]==3) {
				addText("Cat's game!\n");
			} else {
				addText("You lose!\n");
			}
			break;
		}
	}

	// Draw board
	drawBoard();

	// Draw message box
	if (g_MessageBox) {
		g_MessageBox->DrawTrans(Screen->GetBack());
	}

}

// ------------------------------------------------------------------
// cdx_DoFrame - performs drawing of the current frame
// ------------------------------------------------------------------
void cdx_DoFrame()
{
	// read input from devices
	Input->Update();

	// Draw screen
	switch (g_iGameState) {
	case STATE_SPLASH:
		handleSplash();
		break;
	case STATE_MAIN:
		handleMain();
		break;
	}


	if (g_iGameState != STATE_EXIT) {
		Screen->Flip();	 // Flip the back buffer to the front
	}
}



// ------------------------------------------------------------------
// AdjustWinStyle - adjusts the window style according to the mode
// ------------------------------------------------------------------
static void AdjustWinStyle()
{
	RECT        rect = {0, 0, 640, 480};
	DWORD       dwStyle;

	// Change window attributes

	dwStyle = GetWindowStyle(g_hWnd);
	dwStyle &= ~WS_POPUP;
	dwStyle |= WS_OVERLAPPED | WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU;

	SetWindowLong(g_hWnd, GWL_STYLE, dwStyle);

	// Resize the window so that the client area is 640x480

	AdjustWindowRectEx(&rect, GetWindowStyle(g_hWnd), GetMenu(g_hWnd) != NULL,
		GetWindowExStyle(g_hWnd));

	// Just in case the window was moved off the visible area of the
	// screen.

	SetWindowPos(g_hWnd, NULL, 0, 0, rect.right-rect.left,
		rect.bottom-rect.top, SWP_NOMOVE | SWP_NOZORDER |
		SWP_NOACTIVATE);

	SetWindowPos(g_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0,
		SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE);
}



// ------------------------------------------------------------------
// WinProc - handles application messages
// ------------------------------------------------------------------
static long PASCAL WinProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case WM_CLOSE:		cdx_DeInit();

		case WM_DESTROY:	g_iGameState = STATE_EXIT;
							stopServer();
							PostQuitMessage(0);
		                    return 0;
							
		case WM_LBUTTONUP:
							bMouseClicked=true;
							MouseXPos = LOWORD(lParam);  // horizontal position of cursor 
							MouseYPos = HIWORD(lParam);  // vertical position of cursor 
							return 0;

		default:        	return DefWindowProc(hWnd, message, wParam, lParam);
	}
}





// ------------------------------------------------------------------
// InitApp - Create the window and the CDX objects
// ------------------------------------------------------------------
static BOOL InitApp(int nCmdShow)
{
	WNDCLASS WndClass;

	WndClass.style = CS_HREDRAW | CS_VREDRAW;
	WndClass.lpfnWndProc = WinProc;
	WndClass.cbClsExtra = 0;
	WndClass.cbWndExtra = 0;
	WndClass.hInstance = g_hInst;
	WndClass.hIcon = LoadIcon(g_hInst, "APPICON");
	WndClass.hCursor = LoadCursor(0, IDC_ARROW);
	WndClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	WndClass.lpszMenuName = 0;
	WndClass.lpszClassName = szClassName;
	RegisterClass(&WndClass);

	g_hWnd = CreateWindowEx(
		0,
		szClassName,
		szAppName,
 		WS_OVERLAPPED | WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU,
		0,
		0,
		320,
		240,
		NULL,
		NULL,
		g_hInst,
		NULL);

    // when hWnd = -1 there was an error creating the main window
    // CDXError needs a CDXScreen object, if there is none at this early
    // program stage, pass it NULL
	if( !g_hWnd ) 
        CDXError( NULL , "could not create the main window" );


	AdjustWinStyle();

	ShowWindow(g_hWnd, nCmdShow);
	UpdateWindow(g_hWnd);

	return TRUE;
}

// ------------------------------------------------------------------
// WinMain - inital function called by windows
// ------------------------------------------------------------------
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;

	// save the app instance
	g_hInst = hInstance;


	if(!InitApp(nCmdShow)) 
        CDXError( NULL , "could not initialize application" );

	if(!cdx_Init())
	{
		PostQuitMessage(0);
		return FALSE;
	}

	while(1)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if(!GetMessage(&msg, NULL, 0, 0 )) return msg.wParam;
			TranslateMessage(&msg); 
			DispatchMessage(&msg);
		}
		else
		{
			cdx_DoFrame();
		}
	}
}
