// A server for NetTacToe.  The server displays it's output
//  on the screen of the client running on the same computer
// Author : Jimb Esser (wasteland@writeme.com)
// Date : 1-28-1
// Version : 1.0

#include "TTTServer.h"
#include "CDXConnection.h"
#include "TicTacToe.h"
#include "NetTacToe.h"
#include <process.h>
#include <assert.h>

#define IDLE_CYLCE 20 // Number of ms to sleep between polling for input

bool bStopServer=false;
bool bServerRunning=false;

void startServer() {
	_beginthread(tttServerThread, 0, NULL);
}

void stopServer() {
	bStopServer=true;
}

bool bIsServerRunning() {
	return bServerRunning;
}

#define MAX_LEN 1024
void tttServerThread(void * dummy) {
	char data[MAX_LEN];
	int iNumRead, iNodeNum;
	unsigned char iTag;
	CDXConnection con;
	CTicTacToe ttt;
	int x, y, iWin;
	int iCurPlayer=0;
	int iStartingPlayer=0;

	bServerRunning=true;


	addText("[Server] Initializing.\n", true);
	if (FAILED(con.CreateHost())) {
		addText("[Server] Failed to initialize, quitting.\n", true);
		bServerRunning = false;
		return;
	}
	addText("[Server] Listening.\n", true);
	con.Listen(TTT_PORT);
	addText("[Server] Waiting for 2 connections.\n", true);
	con.setMaxConnections(2);
	while (con.getNumConnections()!=2) {
		// Waiting for clients
		Sleep(IDLE_CYLCE);
	}
	addText("[Server] Two connections accepted, starting game.\n", true);

	while (!con.m_bConnectionLost && !bStopServer) {
		// Main loop
		while (!con.isDataReady() &&
			!con.m_bConnectionLost && !bStopServer) {
			// Wait for some data
			Sleep(IDLE_CYLCE);
		}
		if (con.isDataReady()) {
			// Handle the data
			if (FAILED(con.getData(iNodeNum, data, MAX_LEN, iNumRead, &iTag))) {
				addText("[Server] Error receiving data, aborting.\n");
				addText("[Server] Most likely: Other side closed connection.\n");
				bServerRunning = false;
				return;
			}
			switch (iTag) {
			case TAG_CHAT:
				// Send the message to everyone
				if (FAILED(con.sendData(NODE_ALL, data, strlen(data)+1, TAG_CHAT))) {
					addText("[Server] Error sending data, aborting.\n");
					bServerRunning = false;
					return;					
				}
				break;
			case TAG_MARK:
				assert(iNumRead==2);
				if (iNodeNum!=iCurPlayer) {
					// Not their turn!
					strcpy(data, "[Server] Not your turn!\n");
					if (FAILED(con.sendData(iNodeNum, data, strlen(data)+1, TAG_CHAT))) {
						addText("[Server] Error sending data, aborting.\n");
						bServerRunning = false;
						return;					
					}
				} else {
					iCurPlayer = (iCurPlayer + 1) %2;
					x = data[0];
					y = data[1];
					if (ttt.isEmpty(x, y)) { // Only do stuff if it's empty
						ttt.mark(x, y, iNodeNum + 1);
						data[2] = iNodeNum + 1;

						if (FAILED(con.sendData(NODE_ALL, data, 3, TAG_MARK))) {
							addText("[Server] Error sending data, aborting.\n");
							bServerRunning = false;
							return;					
						}

						iWin = ttt.getWinner();
						if (iWin!=0) {
							data[0] = iWin;
							if (FAILED(con.sendData(NODE_ALL, data, 1, TAG_ENDGAME))) {
								addText("[Server] Error sending data, aborting.\n");
								bServerRunning = false;
								return;					
							}

							// Clear the boards of all of them:
							for (x=0; x<3; x++) {
								for (y=0; y<3; y++) {
									data[0]=x;
									data[1]=y;
									data[2]=0;
									ttt.mark(x, y, 0);
									if (FAILED(con.sendData(NODE_ALL, data, 3, TAG_MARK))) {
										addText("[Server] Error sending data, aborting.\n");
										bServerRunning = false;
										return;					
									}
								}
							}
							iStartingPlayer = (iStartingPlayer + 1) %2;
							iCurPlayer = iStartingPlayer;
						}
					}
				}
				break;
			}
		}
	}
}